/*---------------------------------------------------------------------------

    Filename: picturepuzzle.js
    Project:  Picture Puzzle
    (C) Copyright 2008, Kirix Corporation, All Rights Reserved.

    Summary:         Main controls for the Picture Puzzle
    Developer(s):    Jeff Engel

---------------------------------------------------------------------------*/


var g_images = [
    "img/blue.png",
    "img/book.png",
    "img/buds.png",
    "img/cake.png",
    "img/cds.png",
    "img/circle.png",
    "img/dolphin.png",
    "img/dr.png",
    "img/ferriswheel.png",
    "img/flower.png",
    "img/kstrata.png",
    "img/numbers.png",
    "img/fruit.png",
    "img/green.png",
    "img/keyboard.png",
    "img/poinsetta.png",
    "img/rose.png",
    "img/star.png",
    "img/sunset.png",
    "img/wicker.png"
    ];


class PicturePuzzleGame extends Form
{
    function PicturePuzzleGame(x_pos, y_pos, width, height)
    {
        // create the form
        super("Picture Puzzle", x_pos, y_pos, width, height);
        setClientSize(width, height);
        setMinSize(getSize());
        setMaxSize(getSize());


        // top buttons

        m_timeLabel = new Label("00:00:00   ");
        m_timeLabel.setFont(new Font(Font.Default, 20));

        var pictureButton = new BitmapButton(HostBitmap.get("picture", 24),
                                             0, 0, 32, 32);
        pictureButton.click.connect(this, onPictureClicked);

        var reloadButton = new BitmapButton(HostBitmap.get("reload", 24),
                                            0, 0, 32, 32);
        reloadButton.click.connect(this, onShuffleClicked);

        var topLayout = new BoxLayout(Layout.Horizontal);
        topLayout.add(m_timeLabel, 0, Layout.Center);
        topLayout.addStretchSpacer();
        topLayout.add(pictureButton);
        topLayout.addSpacer(2);
        topLayout.add(reloadButton);

        // middle content

        m_tiles = new PicturePuzzleTiles(0, 0, 256);
        m_tiles.setBackgroundColor(SystemColors.ControlFace);
        m_tiles.gameOver.connect(this, onGameOver);

        // main layout

        var vlayout = new BoxLayout(Layout.Vertical);
        vlayout.addSpacer(9);
        vlayout.add(topLayout, 0, Layout.Expand);
        vlayout.addSpacer(10);
        vlayout.add(m_tiles, 0, Layout.Center);

        var layout = new BoxLayout(Layout.Horizontal);
        layout.addStretchSpacer();
        layout.add(vlayout, 0, Layout.Left | Layout.Right, 2);
        layout.addStretchSpacer();

        this.setBackgroundColor(m_bgColor);
        this.setLayout(layout);


        m_timer = new Timer;
        m_timer.setInterval(1000);
        m_timer.tick.connect(this, onTimerTick);


        loadPuzzle(Extension.getBitmapResource("img/kstrata_tiled.png"), true);
    }
    
    function loadPuzzle(img, overrideTile)
    {
        // stop the timer
        m_timer.stop();
        m_timeLabel.setLabel("Loading...");

        // load the new picture
        m_tiles.initGame(img);

        m_timeLabel.setLabel("00:00:00");

        // shuffle the tiles
        m_tiles.shuffle();

        // prepare the game
        m_tiles.mouseLeftDown.connect(this, onInitialTileClick);
    }


    function onInitialTileClick(evt, sender)
    {
        m_tiles.mouseLeftDown.disconnect(this, onInitialTileClick);
        m_startTime = new Date;
        m_timer.start();
    }
    
    function onTimerTick()
    {
        // clear the event queue
        if (!m_timer.isRunning())
            return;

        // calculate the elapsed time
        var elapsed = new Date();
        elapsed = (elapsed.getTime() - m_startTime.getTime()) / 1000;

        var hour = Math.floor(elapsed / 3600);
        hour = hour < 10 ? "0" + hour : hour;
        elapsed -= hour * 3600;

        var min = Math.floor(elapsed / 60);
        min = min < 10 ? "0" + min : min;
        elapsed -= min * 60;

        var sec = elapsed < 10 ? "0" + elapsed : elapsed;

        // display the time
        m_timeLabel.setLabel(hour + ":" + min + ":" + sec);
    }
    
    function onGameOver()
    {
        m_timer.stop();
    }
    
    function onShuffleClicked()
    {
        // stop the timer
        m_timer.stop();
        m_timeLabel.setLabel("00:00:00");

        // shuffle or reset the tiles, depending on if they are in order or not
        if (m_tiles.inOrder())
        {
            m_tiles.shuffle();

            // prepare the game
            m_tiles.mouseLeftDown.connect(this, onInitialTileClick);
        }
        else
        {
            m_tiles.reset();

            m_tiles.mouseLeftDown.disconnect(this, onInitialTileClick);
        }
    }

    function onPictureClicked()
    {
        var path = pictureFlow(g_images);
        if (path)
            loadPuzzle(Extension.getBitmapResource(path));
    }


    // game variables
    var m_status,
        m_timer,
        m_startTime;
    // form controls / layout variables
    var m_img,
        m_bgColor,
        m_tiles,
        m_timeLabel;
}





class PicturePuzzleTiles extends Control
{
    // public events
    var gameOver;

    // game-related variables
    var m_playing,
        m_tiles;

    // shuffle animation variables
    var m_shuffleTimer;

    // tile animation variables
    var m_movingTileTimer,
        m_movingTileBuffer,
        m_movingTilePosition,
        m_movingTileSlope,
        m_explosionTimer;

    // graphics-related variables
    var m_size,
        m_tileSize,
        m_background_color;
    
    // graphics buffers
    var m_tileBuffers,
        m_activeBuffer;

    function PicturePuzzleTiles(x_pos, y_pos, size)
    {
        // since the size of the control will not change, precalculate
        // the overall size (round down to mod-4) and the size of the tiles
        m_size = size - (size % 4);
        m_tileSize = m_size >> 2;

        // create the control
        super(x_pos, y_pos, m_size, m_size);

        // fired when the user has won the game
        gameOver = new Event;

        // event used to trigger a tile move
        m_movingTileTimer = new LimitedTimer;
        m_movingTileTimer.setInterval(10);
        m_movingTileTimer.setMaxTicks(10);
        m_movingTileTimer.tick.connect(this, onTileMove);

        // event used to trigger a tile swap (shuffle)
        m_shuffleTimer = new LimitedTimer;
        m_shuffleTimer.setInterval(150);
        m_shuffleTimer.setMaxTicks(7);
        m_shuffleTimer.tick.connect(this, onShuffle);

        // event used to trigger the tile explosion
        m_explosionTimer = new LimitedTimer;
        m_explosionTimer.setInterval(25);
        m_explosionTimer.setMaxTicks(20);
        m_explosionTimer.tick.connect(this, onExplode);
        
        // triggered when the control is redrawn
        this.paint.connect(this, onPaint);
        this.mouseLeftDown.connect(this, onClick);
    }
    
    function initGame(img, overrideTile)
    {
        // stop the old game
        clearEvents();

        // create the buffers
        initBuffers(img, overrideTile);
        
        // set up the rest of the game
        reset();
    }


    // --- public accessors/mutators ---
    
    function setBackgroundColor(color)
    {
        m_background_color = color;
    }
    
    function getBackgroundColor()
    {
        return m_background_color;
    }
    
    function getSize()
    {
        return new Size(m_size, m_size);
    }
    
    function inOrder()
    {
        for (var i = 0; i < 16; i++)
            if (m_tiles[i] != i)
                return false;
        return true;
    }
    
    function shuffle()
    {
        // stop everything
        clearEvents();

        // redraw the buffer in case an animation was interrupted
        redrawBuffer();

        // start the shuffling
        m_shuffleTimer.start();
    }
    
    function reset()
    {
        // stop everything
        clearEvents();

        // put the tiles in their original order
        m_tiles = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15];

        // draw the buffer with the correct tile positions
        redrawBuffer();
    }


    // --- event handlers ---

    function onPaint(sender, evt)
    {
        // copy the active buffer onto the canvas
        evt.graphics.drawImage(m_activeBuffer, 0, 0, m_size, m_size);
    }

    function onClick(sender, evt)
    {
        // if the game is over or another animation is running don't allow more
        // clicks
        if (!m_playing || m_movingTileTimer.isRunning())
            return;

        // box number that was clicked (0-15)
        var box = Math.floor(evt.x / m_tileSize) +
                  Math.floor(evt.y / m_tileSize) * 4;

        // move up
        if (m_tiles[box - 4] == 0)
        {
            m_tiles[box - 4] = m_tiles[box];
            m_movingTileSlope = new Point(0, -1);
        }
        // move right
        else if (m_tiles[box + 1] == 0)
        {
            m_tiles[box + 1] = m_tiles[box];
            m_movingTileSlope = new Point(1, 0);
        }
        // move down
        else if (m_tiles[box + 4] == 0)
        {
            m_tiles[box + 4] = m_tiles[box];
            m_movingTileSlope = new Point(0, 1);
        }
        // move left
        else if (m_tiles[box - 1] == 0)
        {
            m_tiles[box - 1] = m_tiles[box];
            m_movingTileSlope = new Point(-1, 0);
        }
        // invalid move
        else
            return;

        // set up tile moving information
        var delta = m_tileSize / m_movingTileTimer.getMaxTicks();
        m_movingTileBuffer = m_tileBuffers[m_tiles[box]];
        m_movingTilePosition = new Point(getBoxXCoordinate(box),
                                         getBoxYCoordinate(box));
        m_movingTileSlope.x *= delta;
        m_movingTileSlope.y *= delta;

        // mark the blank box
        m_tiles[box] = 0;

        // start the animation
        m_movingTileTimer.start();
    }
    
    function onTileMove()
    {
        // clear out the event queue
        if (!m_movingTileTimer.isRunning())
            return;

        // clear the position where the tile used to be
        m_activeBuffer.drawImage(m_tileBuffers[0],
                                 Math.round(m_movingTilePosition.x),
                                 Math.round(m_movingTilePosition.y),
                                 m_tileSize, m_tileSize);

        // update the tile's position
        m_movingTilePosition.x += m_movingTileSlope.x;
        m_movingTilePosition.y += m_movingTileSlope.y;
        
        // draw the tile
        m_activeBuffer.drawImage(m_movingTileBuffer,
                                 Math.round(m_movingTilePosition.x),
                                 Math.round(m_movingTilePosition.y),
                                 m_tileSize, m_tileSize);

        // redraw the control
        refresh();

        // when the animation finishes, check if the player won
        if (!m_movingTileTimer.getRemainingTicks() && inOrder())
        {
            // stop the game
            m_playing = false;

            // start the explosion
            m_explosionTimer.start();

            // fire the gameOver event
            gameOver.fire();
        }
    }
    
    function onShuffle()
    {
        // clear out the event queue
        if (!m_shuffleTimer.isRunning())
            return;

        // shuffle all 16 tiles
        for (var i = 0; i < 16; i++)
        {
            var tileNum = m_tiles[i],
                t2 = Math.floor(Math.random() * 16);
            m_tiles[i] = m_tiles[t2];
            m_tiles[t2] = tileNum;
        }

        // refresh the buffer
        redrawBuffer();

        // done shuffling
        if (!m_shuffleTimer.getRemainingTicks())
        {
            // verify the game is solvable
            var inv = 0, row;
            for (var i = 0; i < 16; i++)
            {
                if (!m_tiles[i])
                    row = 4 - (i >> 2);
                else
                    for (var j = i + 1; j < 16; j++)
                        if (m_tiles[j] && m_tiles[j] < m_tiles[i])
                            inv++;
            }

            // start the game if it is solvable
            if ((row & 1) == (inv & 1))
                startPlay();

            // if the game is not solvable, shuffle again
            else
                m_shuffleTimer.tick.fire();
        }
    }
    
    function onExplode()
    {
        // clear out the event queue
        if (!m_explosionTimer.isRunning())
            return;

        // clear the buffer
        m_activeBuffer.drawRectangle(0, 0, m_size, m_size);

        // calculate the offseted center position and frame
        var center = (m_size - m_tileSize) >> 1,
            pct = 1 - 4 * Math.pow(m_explosionTimer.getMaxTicks() -
                                   m_explosionTimer.getRemainingTicks(), 2) /
                                   Math.pow(m_explosionTimer.getMaxTicks(), 2);

        // draw the background text
        if (pct < -0.5)
            m_activeBuffer.drawText("You won!",
                                    (m_size >> 1) - 75,
                                    (m_size >> 1) - 10);

        // copy the tiles onto the buffer according to the animation sequence
        for (var i = 1; i < 16; i++)
            m_activeBuffer.drawImage(m_tileBuffers[m_tiles[i]], center + pct *
                    ((pct < 0 ? center * Math.cos(i * 0.41887902) :
                    getBoxXCoordinate(i) - center)), center + pct * ((pct < 0 ?
                    center * Math.sin(i * 0.41887902) : getBoxYCoordinate(i) -
                    center)), m_tileSize, m_tileSize);

        // refresh the control
        refresh();
    }


    // --- private methods ---

    function initBuffers(img, overrideTile)
    {
        // create the active buffer

        var activeBmp = new Bitmap(m_size, m_size);
        m_activeBuffer = Graphics.fromBitmap(activeBmp);
        m_activeBuffer.setFont(new Font(Font.Modern, 24, Font.Bold));


        // create the tile buffers

        m_tileBuffers = [];
        var fgGr = Graphics.fromBitmap(img),
            tileBmp = new Bitmap(m_tileSize, m_tileSize),
            backgroundPen = new Pen(m_background_color),
            backgroundBrush = new Brush(m_background_color);
        m_activeBuffer.setPen(backgroundPen);
        m_activeBuffer.setBrush(backgroundBrush);

        // create the tiles
        for (var i = 0; i < 16; i++)
        {
            m_tileBuffers[i] = Graphics.fromBitmap(tileBmp);
            m_tileBuffers[i].setPen(backgroundPen);
            m_tileBuffers[i].setBrush(backgroundBrush);

            // don't draw an image on the blank tile
            if (!i)
            {
                m_tileBuffers[i].drawRectangle(0, 0, m_tileSize, m_tileSize);
                continue;
            }

            // copy the image into the tile
            var offset = new Point((i % 4) * m_tileSize,
                                   (i >> 2) * m_tileSize);
            m_tileBuffers[i].drawImage(fgGr,
                          -offset.x,
                          -offset.y,
                          offset.x + m_tileSize,
                          offset.y + m_tileSize);

            // "blank" out the four corners
            m_tileBuffers[i].drawLine(0, 0, 1, 0);
            m_tileBuffers[i].drawLine(m_tileSize - 1, 0, m_tileSize, 0);
            m_tileBuffers[i].drawLine(m_tileSize - 1, m_tileSize - 1,
                                      m_tileSize, m_tileSize - 1);
            m_tileBuffers[i].drawLine(0, m_tileSize - 1, 1, m_tileSize - 1);
        }

        if (overrideTile)
            return;

        // draw the horizontal borders(with light->dark gradients)
        for (var x = 1; x < m_tileSize - 1; x++)
        {
            var c1 = 128 - Math.round(48 * x / m_tileSize);
            var p1 = new Pen(new Color(c1, c1, c1));
            var c2 = 80 - Math.round(48 * x / m_tileSize);
            var p2 = new Pen(new Color(c2, c2, c2));
            
            for (var i = 1; i < 16; i++)
            {
                // top
                m_tileBuffers[i].setPen(p1);
                m_tileBuffers[i].drawLine(x, 0, x + 1, 0);
                if (x == 1)
                    m_tileBuffers[i].drawLine(1, 1, 2, 1);

                // bottom
                m_tileBuffers[i].setPen(p2);
                m_tileBuffers[i].drawLine(x, m_tileSize - 1,
                                          x + 1, m_tileSize - 1);
                if (x == m_tileSize - 2)
                    m_tileBuffers[i].drawLine(x, m_tileSize - 2,
                                              x + 1, m_tileSize - 2);
            }
        }

        // draw the vertical borders (with light->dark gradients)
        for (var y = 1; y < m_tileSize - 1; y++)
        {
            var c1 = 128 - Math.round(48 * y / m_tileSize);
            var p1 = new Pen(new Color(c1, c1, c1));
            var c2 = 80 - Math.round(48 * y / m_tileSize);
            var p2 = new Pen(new Color(c2, c2, c2));
            
            for (var i = 1; i < 16; i++)
            {
                // left
                var c1 = 128 - Math.round(64 * y / m_tileSize);
                m_tileBuffers[i].setPen(p1);
                m_tileBuffers[i].drawLine(0, y, 1, y);
                if (y == m_tileSize - 2)
                    m_tileBuffers[i].drawLine(1, y, 2, y);

                // right
                var c2 = 64 - Math.round(64 * y / m_tileSize);
                m_tileBuffers[i].setPen(p2);
                m_tileBuffers[i].drawLine(m_tileSize - 1, y,
                                          m_tileSize, y);
                if (y == 1)
                    m_tileBuffers[i].drawLine(m_tileSize - 2, 1,
                                              m_tileSize - 1, 1);
            }
        }
    }

    function redrawBuffer()
    {
        // copy the tiles onto the buffer
        for (var i = 0; i < 16; i++)
            m_activeBuffer.drawImage(m_tileBuffers[m_tiles[i]],
                                     getBoxXCoordinate(i),
                                     getBoxYCoordinate(i),
                                     m_tileSize, m_tileSize);
        // refresh the control
        refresh();
    }
    
    function getBoxXCoordinate(num)
    {
        return (num % 4) * m_tileSize;
    }
    
    function getBoxYCoordinate(num)
    {
        return (num >> 2) * m_tileSize;
    }
    
    function clearEvents()
    {
        // stop any current games
        m_playing = false;

        // stop all animations
        m_movingTileTimer.stop();
        m_shuffleTimer.stop();
        m_explosionTimer.stop();
    }
    
    function startPlay()
    {
        // start the game
        m_playing = true;
    }
}
